%{
/*
   

  cisco> show ip accounting
  
  sh$ trafshow

   0.6

  Copyright (c) 1998-2009 may safely be consumed by a BSD or GPL license.
  Written by:   Vladimir Oleynik

   mk_nets() (C) Anatoly A. Orehovsky. AO9-RIPE. AAO1-RIPN
*/

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ac.h"

#if 0
#define YYDEBUG 1
#define YYERROR_VERBOSE
#endif

static int yylex(void);
static int yyparse (void);

static void yyerror(const char *m);

struct IP_AMASK {
	unsigned int ip;
	int automask;
};

struct IP_MASK {
	unsigned int ip;
	unsigned int mask;
};

struct IP_OR_VAR {
	struct IP_MASK *p_ip_mask;
	int n_ip_mask;  /* >=0 number ip:mask, < 0 - var */
	char *var_name;
};

struct L_IP_OR_VAR {
	struct IP_OR_VAR ip_or_var;
	struct L_IP_OR_VAR *next;
};

struct VARS {
	struct L_IP_OR_VAR *value;
	char *var_name;
	int hash;
	int tst_recur;
	struct VARS *next;
};
static struct VARS var_DEBUG = { NULL, "DEBUG",
	(int)'D'+'E'+'B'+'U'+'G', 0, NULL };
static struct VARS * vars = &var_DEBUG;

struct LL_SET {
	double coefficient;
	double pay_level;
};

struct LL_TYPES {
	char *log_level;
	double coefficient;
	double pay_level;
	struct LL_TYPES *next;
};
static struct LL_TYPES *lls;

struct REMS {
	char *rem;
	traf_t sum;
	traf_t all_sum;
	struct LL_TYPES *ll;
	int flg_show;
	traf_t out;
	struct REMS *next;
};
static struct REMS *rems;

struct CONF_LINE {
	struct IP_MASK ip1;
	struct IP_MASK ip2;
	struct REMS *rem;
	int flg_debug_line;
	struct CONF_LINE *next;
};
static struct CONF_LINE *cl, *cl_prev;

#define TOKEN_MODE     0 /* any word is token */
#define ALWAYS_WORD    1 /* any word is WORD always */
#define ALWAYS_IP      2 /* any word is IP */
#define ALWAYS_NUMBER  3 /* any word is is NUMBER or DNUMBER if have one dot */
static int lex_stat = TOKEN_MODE;  /* default */

/*
   special inet_aton()
   return mask for short and full IP form:
   returned 8  for 1       or 1.
   returned 16 for 1.2     or 1.2.
   returned 24 for 1.2.3   or 1.2.3.
   returned 32 for 1.2.3.4
   returned -8 or -16 or -24 or -32 for octet is not number or biggest of 255
   returned -40 for non zero terminated: 1.2.3.4[^\0]
   IP value stored in local endian
*/
static int
scan_ip_short_form (const char *p, unsigned int *ip)
{
  int auto_mask = 8;
  int j;

  *ip = 0;
  for (j = 0; j < 4; j++) {
	unsigned int octet;

	if ((*p < '0' || *p > '9') && (*p != '\0' || j == 0))
	  return -auto_mask;
	octet = 0;
	while (*p >= '0' && *p <= '9') {
	  octet *= 10;
	  octet += *p - '0';
	  if (octet > 255)
		return -auto_mask;
	  p++;
	}
	if (*p == '.' && j != 3)
	  p++;
	if (*p != '\0')
	  auto_mask += 8;
	*ip = ((*ip) << 8) | octet;
  }
  if (*p != '\0')
	return -auto_mask;
  return auto_mask;
}


static void make_ip(struct IP_AMASK *ip1, int delim, struct IP_AMASK *ip2, struct IP_OR_VAR *ret);
static void make_ip_var(const char *var, struct IP_OR_VAR *ret);
static struct L_IP_OR_VAR *make_iplist(struct IP_OR_VAR ip, struct L_IP_OR_VAR *l);
static void make_var(char *var, struct L_IP_OR_VAR *vlist);
static char *add_word_to_list_words(char *l, const char *w);
static void set_ll(char *log_level, struct LL_SET *ll_set);
static void conf_line(struct IP_OR_VAR *ip1, struct IP_OR_VAR *ip2, void *sll,
				      char *rem, int rl);

static char *start_line;
static long skip1, skip2;

static int eof_any;
#define CONF_EOF  0
#define CONF_ANY  1
#define CONF_EOL  2
#define CONF_NONE 3
static char *stop_line;

static int flg_recount;

#define SUM_NO   0
#define SUM_YES  1
#define SUM_ONLY 2
static int flg_sum = SUM_YES;

#define SORT_NO   0
#define SORT_YES  1
#define SORT_LL   2
static int flg_sort = SORT_YES;

#define DEBUG_NO   0
#define DEBUG_YES  1
static int flg_debug;
%}

%start list

%union  {
	long lval;
	double dval;
	char * sval;
	struct IP_AMASK ip_amask;
	struct IP_OR_VAR ip_or_var;
	struct L_IP_OR_VAR *lip_or_var;
	struct LL_SET ll_set;
	}

%token CFG_EOF
%token <lval> NUMBER
%token <dval> DNUMBER
%token START SKIP1 SKIP2 END RECOUNT SUM SORT DEBUG LL VAR
%token CONF_IP
%token <sval> STRING WORD
%token <ip_amask> IP32 IP
%token Eof EOL NONE YES NO ONLY

%type <ip_or_var>  ip
%type <lip_or_var> iplist
%type <sval> list_words rem
%type <ll_set> ll


%%
list:
	stat '\n'
	| list stat '\n'
	;
stat:
	START '=' STRING        { start_line = $3; }
	| VAR { lex_stat = ALWAYS_WORD; } WORD { lex_stat = ALWAYS_IP; } '{' can_cr iplist '}' { make_var($3, $7); lex_stat = TOKEN_MODE; }
	| SKIP1 '=' { lex_stat = ALWAYS_NUMBER; } NUMBER   { skip1=$4; lex_stat = TOKEN_MODE; }
	| SKIP2 '=' { lex_stat = ALWAYS_NUMBER; } NUMBER   { skip2=$4; lex_stat = TOKEN_MODE; }
	| END '=' end
	| RECOUNT '=' recount
	| SUM '=' sum
	| SORT '=' sort
	| DEBUG '=' debug
	| LL '=' { lex_stat = ALWAYS_WORD; } WORD { lex_stat = ALWAYS_NUMBER; } ll  { set_ll($4, &$6); lex_stat = TOKEN_MODE; }
	| CONF_IP '=' '{' '\n' { lex_stat = ALWAYS_IP; } conflines '}' { lex_stat = TOKEN_MODE; }
	| CFG_EOF              { return 0; }
	;
end:
	STRING          { eof_any = CONF_ANY; stop_line = $1; }
	| Eof           { eof_any = CONF_EOF; }
	| EOL           { eof_any = CONF_EOL; }
	| NONE          { eof_any = CONF_NONE; }
	;
recount:
	YES             { flg_recount = 1; }
	| NO            { flg_recount = 0; }
	;
sum:
	YES             { flg_sum = SUM_YES; }
	| NO            { flg_sum = SUM_NO; }
	| ONLY          { flg_sum = SUM_ONLY; }
	;
sort:
	YES             { flg_sort = SORT_YES; }
	| NO            { flg_sort = SORT_NO; }
	| LL            { flg_sort = SORT_LL; }
	;
debug:
	YES             { flg_debug = DEBUG_YES; }
	| NO            { flg_debug = DEBUG_NO; }
	;
ll:
	NUMBER NUMBER           { $$.coefficient = $1; $$.pay_level = $2; }
	| DNUMBER NUMBER        { $$.coefficient = $1; $$.pay_level = $2; }
	| NUMBER DNUMBER        { $$.coefficient = $1; $$.pay_level = $2; }
	| DNUMBER DNUMBER       { $$.coefficient = $1; $$.pay_level = $2; }
	;
iplist:
	ip can_cr               { $$ = make_iplist($1, NULL); }
	| iplist ip can_cr      { $$ = make_iplist($2, $1); }
	;
ip:
	IP32            { make_ip(&$1, '\0', &$1, &$$); }
	| IP '/' IP32   { make_ip(&$1, '/', &$3, &$$); }
	| IP '-' IP32   { make_ip(&$1, '-', &$3, &$$); }
	| '$' { lex_stat = ALWAYS_WORD; } WORD { lex_stat = ALWAYS_IP; make_ip_var($3, &$$); }
	;
can_cr:
	| '\n'
	;
confline:
	ip ip { lex_stat = ALWAYS_WORD; } WORD rem '\n' { conf_line(&$1, &$2, $4, $5, 0); lex_stat = ALWAYS_IP; }
	;
conflines:
	confline
	| conflines confline
	;
rem:
	STRING
	| list_words
	;
list_words:
	WORD
	| list_words WORD       { $$ = add_word_to_list_words($1, $2); }
	;
%%


static int empty_line = 1;

struct TOKENS {
	const char *token;
	int tok;
};
static struct TOKENS const tokens[] = {
	{ "START", START },
	{ "SKIP1", SKIP1 },
	{ "SKIP2", SKIP2 },
	{ "END", END },
	{ "RECOUNT", RECOUNT },
	{ "SUM", SUM },
	{ "SORT", SORT },
	{ "DEBUG", DEBUG },
	{ "LL", LL },
	{ "VAR", VAR },
	{ "CONF_IP", CONF_IP },
	{ "EOF", Eof },
	{ "EOL", EOL },
	{ "NONE", NONE },
	{ "YES", YES },
	{ "NO", NO },
	{ "ONLY", ONLY },
	{ NULL, 0 }
};

static char *b_in, *b_in0;
static size_t fsz;
static size_t cur_fsz;
static int linenum;

static int getchar0(int flg_Unexpected_EOF)
{
	int c;

	if(cur_fsz == fsz) {
		if(flg_Unexpected_EOF) {
			fprintf(stderr, "%d: unexpected EOF\n", linenum);
			exit(3);
		}
		b_in[--cur_fsz] = '\n';
		return EOF;
	}
	c = b_in[cur_fsz++];
	if(c == '\0') {
		fprintf(stderr, "%d: unexpected zero byte\n", linenum);
		exit(3);
	}
	return c;
}
#define ungetc()        cur_fsz--


static void __attribute__((noreturn)) err_synt_num(const char *s)
{
	fprintf(stderr, "%d: bad number syntax: %s\n", linenum, s);
	exit(4);
}

static int detected_cr = 1; /* count '\n' after the syntax analysis */
static int yylex(void)
{
    int c;
    char *s;
    int auto_mask;
    unsigned int ip;

#if YYDEBUG
    {
	static const char *s_lex_stat[] = {
				"TOKEN_MODE",
				"ALWAYS_WORD",
				"ALWAYS_IP",
				"ALWAYS_NUMBER"
	};
	fprintf(stderr, "---> called yylex(void) lex_stat=%s\n", s_lex_stat[lex_stat]);
    }
#endif
    while(1) {
	if(detected_cr) {
		detected_cr = 0;
		linenum++;
	}
	c = getchar0(0);
	if(c == EOF)
		return CFG_EOF;
	if(c == ' ' || c == '\t' || c == '\v')
		continue;
	if(c == '\n') {
		detected_cr++;
		if(empty_line)
			continue;
		empty_line = 1;
#if YYDEBUG
		fprintf(stderr, "---> yylex() return '\\n'\n");
#endif
		return c;

	}
	if(c == '#' && empty_line) {    /* ^[ \t\v]*#.*$ */
		while(1) {
			c = getchar0(0);
			if(c == EOF)
				exit(0);
			if(c == '\n') {
				detected_cr++;
				break;
			}
		}
		continue;
	}
	empty_line = 0;
	if(strchr("$-/={}", c) && lex_stat != ALWAYS_WORD) {
#if YYDEBUG
		fprintf(stderr, "---> yylex() return '%c'\n", c);
#endif
		return c;
	}
	if(c == '"') {
		yylval.sval = b_in0;
		while(1) {
			c = getchar0(1);
			if(c == '"')
				break;
			*b_in0++ = c;
			if(c == '\n')
				linenum++;
		}
		*b_in0++ = '\0';
#if YYDEBUG
		fprintf(stderr, "---> yylex() return STRING\n");
#endif
		return STRING;
	}

	s = b_in0;
	while(1) {
		*b_in0++ = c;
		c = getchar0(1);
		if(strchr("\t\n\v ", c) != NULL) {
			ungetc();
			break;
		}
		if(lex_stat != ALWAYS_WORD && strchr("$-/={}", c) != NULL) {
			ungetc();
			break;
		}
	}
	*b_in0 = '\0';  /* but saved in 'c' */
	if(lex_stat == TOKEN_MODE) {
		int i;
		for(i = 0; tokens[i].token != NULL; i++)
			if(strcasecmp(s, tokens[i].token) == 0) {
#if YYDEBUG
				fprintf(stderr, "---> yylex() return TOKEN=%s\n", tokens[i].token);
#endif
				*b_in0 = c;     /* restore delimiter */
				return tokens[i].tok;

			}
		fprintf(stderr, "%d: undefined reserved word: %s\n", linenum, s);
		exit(3);
	}
	/* do not restore b_in0: first a word must be the TOKEN in syntax */
	b_in0++;
	if(lex_stat == ALWAYS_WORD) {
#if YYDEBUG
		fprintf(stderr, "---> yylex() return WORD=%s\n", s);
#endif
		yylval.sval = s;
		return WORD;
	}


	if(lex_stat == ALWAYS_IP) {
		auto_mask = scan_ip_short_form(s, &ip);
		if(auto_mask < 0) {
			fprintf(stderr, "%d: bad ip syntax: %s\n", linenum, s);
			exit(4);
		}
		yylval.ip_amask.ip = ip;
		yylval.ip_amask.automask = auto_mask;
#if YYDEBUG
		fprintf(stderr, "---> yylex() ip delim='%c'\n", c);
#endif
		if(c != '-' && c != '/') {
#if YYDEBUG
			fprintf(stderr, "---> yylex() return IP32=%s\n", s);
#endif
			return IP32;
		}
#if YYDEBUG
		fprintf(stderr, "---> yylex() return IP=%s\n", s);
#endif
		return IP;
	}

	/* lex_stat == ALWAYS_NUMBER */
	if(strchr(s, '.') == NULL) {
		char *e;

		errno = 0;
		yylval.lval = strtol(s, &e, 10);
		if(*e != '\0' || errno != 0)
			err_synt_num(s);
#if YYDEBUG
		fprintf(stderr, "---> yylex() return NUMBER=%s\n", s);
#endif
		return NUMBER;
	} else {
		char *e;

		errno = 0;
		yylval.dval = strtod(s, &e);
		if(*e != '\0' || errno != 0)
			err_synt_num(s);
#if YYDEBUG
		fprintf(stderr, "---> yylex() return DNUMBER=%s\n", s);
#endif
		return DNUMBER;
	}
    }
}

static void yyerror(const char *m)
{
	fprintf(stderr, "%d: %s\n", linenum, m);
	exit(5);
}

static int mk_nets_count;
static void
mk_nets (unsigned int f, unsigned int l, struct IP_MASK *ip_mask)
{
  unsigned int mask;

  mask = 0xffffffffL;

  while (((f & mask) != (l & mask)) && ((f & (mask << 1)) == f) &&
		 ((f | ~(mask << 1)) <= l))
	mask <<= 1;
  if (ip_mask) {
	ip_mask[mk_nets_count].ip = f & mask;
	ip_mask[mk_nets_count].mask = mask;
  }
  mk_nets_count++;
  if ((f & mask) != (l & mask))
	mk_nets ((f | ~mask) + 1, l, ip_mask);

}

static void *xmalloc(size_t size)
{
	void *r = malloc(size);
	if(r == NULL) {
		fprintf(stderr, "Memory exhausted while parse %d line\n", linenum);
		exit(100);
	}
	return r;
}

static void
make_ip(struct IP_AMASK *ip1, int delim, struct IP_AMASK *ip2, struct IP_OR_VAR *ret)
{
	switch(delim) {
		case '\0':
			ret->p_ip_mask = xmalloc(sizeof(struct IP_MASK));
			ret->n_ip_mask = 1;
			ret->p_ip_mask->ip = ip1->ip;
			ret->p_ip_mask->mask =
				ip1->automask == 32 ? 0xffffffff :
				( ~ (0xffffffffUL >> ip1->automask));
			break;
		case '/':
			if(ip2->automask != 32 && ip2->automask != 8)
				yyerror("strange mask");
			ret->p_ip_mask = xmalloc(sizeof(struct IP_MASK));
			ret->n_ip_mask = 1;
			ret->p_ip_mask->ip = ip1->ip;
			if(ip2->automask == 32) {
				ret->p_ip_mask->mask = ip2->ip;
			} else {
				unsigned m = ip2->ip >> 24;
				if(m > 32) {
					yyerror("strange mask bits");
				}
				ret->p_ip_mask->mask =
					m == 32 ? 0xffffffff :
					( ~ (0xffffffffUL >> m));
			}
			break;
		case '-':
			if (ip1->ip > ip2->ip)
				yyerror("strange IP range");
			mk_nets_count = 0;
			mk_nets (ip1->ip, ip2->ip, NULL);
			if (mk_nets_count <= 0)
				yyerror("strange IP range");
			ret->p_ip_mask = xmalloc(mk_nets_count * sizeof(struct IP_MASK));
			ret->n_ip_mask = mk_nets_count;
			mk_nets_count = 0;
			mk_nets (ip1->ip, ip2->ip, ret->p_ip_mask);
			break;
	}
}

static struct VARS *tst_var_exist(const char *var, int flg_error)
{
	struct VARS *cur;
	int hash;
	const char *v;

	if(*var == '\0' || *var == '$') {
		fprintf(stderr, "%d: strange variable name\n", linenum);
		exit(23);
	}

	hash = 0;
	v = var;
	while(*v != '\0') {
		hash += (unsigned char)*v;
		v++;
	}
	for(cur = vars; cur != NULL; cur = cur->next) {
		if(cur->hash == hash && strcmp(var, cur->var_name) == 0)
			break;
	}
	if(flg_error) {
	  if(cur == NULL) {
		fprintf(stderr, "%d: variable %s undefined\n", linenum, var);
		exit(12);
	  }
	  if(cur == &var_DEBUG) {
		fprintf(stderr, "%d: variable %s reserved\n", linenum, var);
		exit(12);
	  }
	}
	return cur;
}

static void
make_ip_var(const char *var, struct IP_OR_VAR *ret)
{
	struct VARS *cur = tst_var_exist(var, 1);

	ret->var_name = cur->var_name;
	ret->n_ip_mask = -1;
}

static void
make_var(char *var, struct L_IP_OR_VAR *vlist)
{
	struct VARS *cur = tst_var_exist(var, 0);
	int hash;

	if(cur) {
		fprintf(stderr, "%d: variable %s defined already\n", linenum, var);
		exit(12);
	}
	cur = xmalloc(sizeof(struct VARS));
	cur->value = vlist;
	cur->var_name = var;
	hash = 0;
	while(*var != '\0') {
		hash += (unsigned char)*var;
		var++;
	}
	cur->hash = hash;
	cur->next = vars;
	vars = cur;
}

static struct L_IP_OR_VAR *
make_iplist(struct IP_OR_VAR ip, struct L_IP_OR_VAR *l)
{
	struct L_IP_OR_VAR *ret = xmalloc(sizeof(struct L_IP_OR_VAR));

	ret->ip_or_var = ip;
	ret->next = NULL;
	if(l == NULL) {
		return ret;
	} else {
		struct L_IP_OR_VAR *prev;

		for(prev = l; prev->next != NULL; prev = prev->next)
			;
		prev->next = ret;
		return l;
	}
}

static char *add_word_to_list_words(char *l, const char *w)
{
	size_t len = strlen(l);

	l[len] = ' ';
	strcpy(l + len + 1, w);
	return l;
}

static void set_ll(char *log_level, struct LL_SET *ll_set)
{
	struct LL_TYPES *ret = xmalloc(sizeof(struct LL_TYPES));

	ret->log_level = log_level;
	ret->coefficient = ll_set->coefficient;
	ret->pay_level = ll_set->pay_level;
	ret->next = lls;
	lls = ret;
}

static volatile int n_expand;
static void expand_recur(struct VARS *v, struct IP_MASK *ip)
{
	struct L_IP_OR_VAR *value = v->value;

	if(v->tst_recur) {
		fprintf(stderr, "%d: the recur use %s\n", linenum, v->var_name);
		exit(19);
	}
	v->tst_recur = 1;
	while(value != NULL) {
		if(value->ip_or_var.n_ip_mask < 0) {
			expand_recur(
				tst_var_exist(value->ip_or_var.var_name, 1),
				ip);
		} else {
			int i = value->ip_or_var.n_ip_mask;
			if(ip) {
				int j;

				for(j = 0; j < i; j++)
				  ip[n_expand++] = value->ip_or_var.p_ip_mask[j];
			} else {
				n_expand += i;
			}
		}
		value = value->next;
	}
}

static struct IP_MASK *expand(struct VARS *v, int *n_ip)
{
	struct VARS *cur;
	struct IP_MASK *ip;

	n_expand = 0;
	for(cur = vars; cur != NULL; cur = cur->next)
		cur->tst_recur = 0;
	expand_recur(v, NULL);
	*n_ip = n_expand;
	ip = xmalloc(sizeof(struct IP_MASK) * n_expand);
	n_expand = 0;
	for(cur = vars; cur != NULL; cur = cur->next)
		cur->tst_recur = 0;
	expand_recur(v, ip);
	return ip;
}

static char *
change_rem(const char *rem, char *start, const char *end, const char *chg)
{
	char *r;
	size_t lr, lc;

	*start = '\0';
	lr = start - rem;
	lc = strlen(chg);
	r = xmalloc(lr + lc + strlen(end) + 1);
	memcpy(r, rem, lr);
	memcpy(r + lr, chg, lc);
	strcpy(r + lr + lc, end);
	*start = '$';
	return r;
}

static int max_length_type;

static void
conf_line(struct IP_OR_VAR *ip1, struct IP_OR_VAR *ip2, void *sll, char *rem,
		int rl)
{
	struct VARS *v1 = NULL, *v2 = NULL;
	struct VARS *vr1 = NULL;
	struct LL_TYPES *ll;
	char *s;
	char *start = NULL, *end;
	struct IP_MASK *sip1, *sip2;
	struct L_IP_OR_VAR *value;
	int flg_debug_line = flg_debug;

#if YYDEBUG
	fprintf(stderr, "conf_line(rem=%s, rl=%d)\n", rem, rl);
#endif
	if(rl > 2) {
		fprintf(stderr, "%d: too many used variable in remarks\n", linenum);
		exit(21);
	}
	if(ip1->n_ip_mask < 0)
		v1 = tst_var_exist(ip1->var_name, 1);
	if(ip2->n_ip_mask < 0)
		v2 = tst_var_exist(ip2->var_name, 1);
	if(rl == 0) {
		for(ll = lls; ll != NULL; ll = ll->next) {
			if(strcmp(ll->log_level, (char *)sll) == 0)
				break;
		}
		if(ll == NULL) {
			static struct LL_SET zero_pay_ll;

			set_ll((char *)sll, &zero_pay_ll);
			ll = lls;
		}
	} else {
		ll = sll;
	}
	for(s = rem; *s != '\0'; ) {
		if(*s == '$') {
			char delim;

			end = s + 1;
			while(*end != '\0' && *end != ' ' && *end != '\t')
				end++;
			delim = *end;
			*end = '\0';
			vr1 = tst_var_exist(s + 1, 0);
			*end = delim;
			if(vr1 == &var_DEBUG) {
				flg_debug_line++;
				/* remove $DEBUG from remark */
				for(start = s ; ; end++) {
					*s++ = *end;
					if(*end == '\0')
						break;
				}
				s = start;
				start = NULL;
				continue;
			} else if(vr1 != NULL && (vr1 == v1 || vr1 == v2)) {
				start = s;
#if YYDEBUG
				fprintf(stderr, "conf_line(): vr1->var_name=%s\n", vr1->var_name);
#endif
				break;
			}
			s = end;
			continue;
		}
		s++;
	}
	/* count real lines - expand vars */
	if(start == NULL) {
		/* full expand */
		int col1, col2, i, j;
		struct CONF_LINE *c;
		struct REMS *r_cur;

		if(v1) {
			sip1 = expand(v1, &col1);
		} else {
			col1 = ip1->n_ip_mask;
			sip1 = ip1->p_ip_mask;
		}
		if(v2) {
			sip2 = expand(v2, &col2);
		} else {
			col2 = ip2->n_ip_mask;
			sip2 = ip2->p_ip_mask;
		}
#if YYDEBUG
		fprintf(stderr, "conf_line(): simple %d*%d rem=%s\n", col1, col2, rem);
#endif
		c = xmalloc(sizeof(struct CONF_LINE) * (col1 * col2));
		r_cur = xmalloc(sizeof(struct REMS));
		r_cur->rem = rem;
		i = s - rem;    /* strlen(rem) */
		if(max_length_type < i)
			max_length_type = i;
		r_cur->sum = 0;
		r_cur->all_sum = 0;
		r_cur->out = 0;
		r_cur->ll = ll;
		r_cur->flg_show = 0;
		r_cur->next = rems;
		rems = r_cur;
		for(i = 0; i < col1; i++)
			for(j = 0; j < col2; j++) {
				c->ip1 = sip1[i];
				c->ip2 = sip2[j];
				c->rem = r_cur;
				c->flg_debug_line = flg_debug_line;
				c->next = NULL;
				if(cl == NULL) {
					cl = cl_prev = c;
				} else {
					cl_prev->next = c;
					cl_prev = c;
				}
				c++;
			}
		if(v2)
			free(sip2);
		if(v1)
			free(sip1);
		return;
	}
	/* recur call with 1-level expand first var in remark */
	if(vr1 == v2)
		ip2 = NULL;
	 else
		ip1 = NULL;

	for(value = vr1->value; value != NULL; value = value->next) {
		if(value->ip_or_var.n_ip_mask < 0)
			s = change_rem(rem, start, end, value->ip_or_var.var_name);
		else
			s = change_rem(rem, start, end, vr1->var_name);
		conf_line(ip1 != NULL ? ip1 : &value->ip_or_var,
			  ip2 != NULL ? ip2 : &value->ip_or_var,
			  ll, s, rl + 1);
	}
}

static void __attribute__((noreturn)) cf(void)
{
	fprintf (stderr, "%d: corrupted traffic file?\n", linenum);
	exit(6);
}

static char *
skip_word (char **ep)
{
  char *p, *r;

  p = *ep;
  if(p == NULL)
	cf();

  while (*p == ' ' || *p == '\t')
	p++;
  r = p;
  while (*p != ' ' && *p != '\t') {
	if (*p == '\0') {
	  *ep = NULL;
	  return r;
	}
	p++;
  }
  *p = '\0';
  *ep = p+1;
  return r;
}

static void
fmt_out (traf_t tbd)
{
  int gig;
  int meg;
  int kb;
  int b;
  double bd = tbd, bg;

  gig = bd / 1000.0 / 1000.0 / 1000.0;
  bg = ((double) gig) * 1000.0 * 1000.0 * 1000.0;
  meg = ((long) (bd / 1000.0 / 1000.0)) - gig * 1000;
  kb =
	((long) (bd / 1000.0)) - ((long) meg) * 1000 - ((long) gig) * 1000 * 1000;
  b = bd - bg - ((double) kb) * 1000.0 - ((double) meg) * 1000.0 * 1000.0;
  if (gig)
	printf ("%3d %03d %03d.%03d", gig, meg, kb, b);
  else {
	printf ("    ");
	if (meg)
	  printf ("%3d %03d.%03d", meg, kb, b);
	else {
	  printf ("    ");
	  printf ("%3d.%03d", kb, b);
	}
  }
}

struct OUT_PAY {
    traf_t traf;
    double pay;
    const char *type;
};

static int
qsort_compar_pay (const void *p1, const void *p2)
{
  struct OUT_PAY *inf_s1;
  struct OUT_PAY *inf_s2;
  double df;

  inf_s1 = (struct OUT_PAY *) p1;
  inf_s2 = (struct OUT_PAY *) p2;
  df = inf_s2->pay - inf_s1->pay;
  if (df != 0)
	return df < 0 ? -1 : 1;
  return 0;
}

static int
qsort_compar_traf (const void *p1, const void *p2)
{
  struct OUT_PAY *inf_s1;
  struct OUT_PAY *inf_s2;
  traf_t df;

  inf_s1 = (struct OUT_PAY *) p1;
  inf_s2 = (struct OUT_PAY *) p2;
  df = inf_s2->traf - inf_s1->traf;
  if (df != 0)
	return df < 0 ? -1 : 1;
  return 0;
}


static void
v_cur (int flg_end)
{
  int j, n;
  struct REMS *cur;
  struct OUT_PAY *nr;

  if (flg_end && flg_sum != SUM_NO)
	printf ("-----------------------------------------------\n");
  else if (!flg_end && flg_sum != SUM_ONLY)
	printf ("--\n");
  else if (!flg_end && flg_sum == SUM_ONLY)
	return;

  n = 0;
  for(cur = rems; cur; cur = cur->next) {
    if (!flg_end && cur->sum) {
	cur->all_sum += cur->sum;
	cur->out = cur->sum;
	n++;
    } else if (flg_end && (cur->all_sum += cur->sum)) {
	cur->out = cur->all_sum;
	n++;
    }
  }
  if(flg_sort == SORT_NO) {
    for(cur = rems; cur; cur = cur->next) {
	if (cur->out) {
		fmt_out (cur->out);
		printf (" %*s %10.2f\n", -max_length_type, cur->rem,
			cur->ll->coefficient * cur->out * cur->ll->pay_level);
	}
    }
  } else {
    nr = malloc (n * sizeof (struct OUT_PAY));
    if (nr == NULL) {
	fprintf (stderr, "Not enough memory for sorting\n");
	exit(101);
    }
    n = 0;
    for(cur = rems; cur; cur = cur->next) {
      if (cur->out) {
	  nr[n].traf = cur->out;
	  nr[n].pay = cur->ll->coefficient * cur->out * cur->ll->pay_level;
	  nr[n].type = cur->rem;
	  n++;
      }
    }
    qsort (nr, n, sizeof (struct OUT_PAY), flg_sort == SORT_LL ?
				qsort_compar_pay : qsort_compar_traf);
    for (j = 0; j < n; j++) {
	fmt_out (nr[j].traf);
	printf (" %*s %10.2f\n", -max_length_type, nr[j].type, nr[j].pay);
    }
    free(nr);
  }

  if(!flg_end)
    for(cur = rems; cur; cur = cur->next) {
	cur->sum = 0;
	cur->out = 0;
    }
}

/* fast inet_aton(), IP value stored in local endian */
static int
scan_ip (const char *p, unsigned int *pip)
{
  const char *o;
  unsigned int ip = 0;
  static const char delims[] = "...";
  const char *pd = delims;

  o = p;
  while(1) {
	unsigned int octet = 0;
	while (*o >= '0' && *o <= '9') {
		octet *= 10;
		octet += *o - '0';
		if (octet > 255)
			return octet;
		o++;
	}
	ip = (ip << 8) | octet;
	if(o == p || *o != *pd)
		return 1;
	if(*pd == '\0')
		break;
	pd++;
	o++;
	p = o;
  }
  *pip = ip;
  return 0;
}

static FILE *cfg;
static char str_buf[1024];
static void parse_traff(void)
{
	char *p, *s, *wip1, *wip2;
	int started = 0;
	int n;
	unsigned int ip1, ip2;
	struct CONF_LINE *c;

	linenum = 1;
	while (fgets (str_buf, sizeof(str_buf), cfg) != NULL) {
		p = strchr(str_buf, '\n');
		if(p == NULL) {
			fprintf(stderr, "traffic file line %d too long\n", linenum);
			exit(24);
		}
		*p = '\0';
		s = str_buf;
		while((p = strchr(s, '\r')) != NULL) {
			char c1;
			s = p;
			do {
				c1 = p[1];
				*p++ = c1;
			} while(c1 != '\0');
		}
		p = str_buf;
		if (start_line == NULL) {
			started++;
		} else if ((eof_any == CONF_NONE || started == 0) &&
			   !strncmp (start_line, p, strlen(start_line))) {
			started++;
			if (eof_any == CONF_NONE) {
				if (flg_recount == 0)
					break;
				v_cur (0);
			}
			continue;
		}
		if(!started)
			continue;
		if ((eof_any == CONF_ANY && !strncmp (stop_line, p, strlen(stop_line)))
				|| (eof_any == CONF_EOL && *p == '\0')) {
			if (flg_recount) {
				started = 0;
				v_cur (0);
				continue;
			}
			break;
		}
		s = skip_word(&p);
		wip1 = s;
		if (scan_ip (s, &ip1))          /* From */
			cf();
		for(n = 0; n < skip1; n++)
			skip_word(&p);
		s = skip_word(&p);
		wip2 = s;
		if (scan_ip (s, &ip2))          /* To */
			cf();
		for(c = cl; c != NULL; c = c->next) {
			if ((ip1 & c->ip1.mask) == c->ip1.ip &&
					(ip2 & c->ip2.mask) == c->ip2.ip) {
				if(c->rem->flg_show) {
					traf_t traf;

					for(n = 0; n < skip2; n++)
						skip_word(&p);
					s = skip_word(&p);
					errno = 0;
					traf = STRTO(s, &p, 10);
					if (errno != 0 || *p != '\0' || p == s)
						cf();
					c->rem->sum += traf;
					if (c->flg_debug_line)
						printf ("%-15s %-15s %*s %16.0f\n", wip1, wip2,
							-max_length_type, c->rem->rem,
							(double)traf);
				} else {
				  if (c->flg_debug_line)
					printf ("%-15s %-15s %s\n", wip1, wip2, c->rem->rem);
				}
				break;
			}
		}
		if (c == NULL)
			printf ("%.15s %.15s (%d: ?Not configured?)\n", wip1, wip2, linenum);
		linenum++;
	}
	if (flg_recount)
		v_cur (0);
	if (flg_sum != SUM_NO)
		v_cur (1);
}

#include <sys/stat.h>

int main(int argc, char **argv)
{
  struct stat st;
  char *a0, *s;

  if (argc < 3) {
usage:
	a0 = argv[0];
	for (s = a0; *s; )
		if (*(s++) == '/') a0 = s;

	fprintf (stderr, "Usage: %s config_file traffic_file [log-levels]\n",
			 a0);
	return 1;
  }

  cfg = fopen (argv[1], "r");
  if (cfg == NULL) {
	fprintf (stderr, "Can`t open config file `%s': %m.\n", argv[1]);
	goto usage;
  }
  if(fstat(fileno(cfg), &st)) {
	fprintf (stderr, "Can`t stat config file `%s': %m.\n", argv[1]);
	goto usage;
  }
  fsz = (size_t)st.st_size;
  if((off_t)fsz != st.st_size) {
	fprintf (stderr, "File `%s' too long.\n", argv[1]);
	return 2;
  }
  b_in0 = b_in = malloc(fsz);
  if(b_in == NULL) {
	fprintf (stderr, "Can`t malloc memory for config file %s: %m.\n", argv[1]);
	return 3;
  }
  if(fread(b_in, 1, fsz, cfg) != fsz) {
	fprintf (stderr, "Can`t load config file `%s' in memory: %m.\n", argv[1]);
	return 3;
  }
  fclose(cfg);
#if YYDEBUG
  yydebug = 1;
#endif
  yyparse();
  if(cl == NULL) {
	fprintf (stderr, "Can`t found main config lines for counting\n");
	return 4;
  }
  cfg = fopen (argv[2], "r");
  if (cfg == NULL) {
	fprintf (stderr, "Can`t open traffic file `%s': %m.\n", argv[2]);
	goto usage;
  }

  if (argc >= 4) {
	struct REMS *r_cur;
	char **shows = argv+3;
	for(shows = argv+3; *shows != NULL; shows++) {
	  int found = 0;
	  for(r_cur = rems; r_cur != NULL; r_cur = r_cur->next) {
	    if(!r_cur->flg_show && strcmp(r_cur->ll->log_level, *shows) == 0) {
		found = 1;
		r_cur->flg_show = 1;
	    }
	  }
	  if(found == 0)
	    fprintf (stderr, "Warning: can not found configured %s log-level\n", *shows);
	}
  } else {
	struct REMS *r_cur;

	for(r_cur = rems; r_cur != NULL; r_cur = r_cur->next)
		r_cur->flg_show = 1;
  }
  parse_traff();

  return 0;
}
